/**
 * The type of arguments of the `escapeValues` and `escape` helpers.
 */
export type EscapeMap = Map<any, (value?: any) => string>;

/**
 * The type of arguments of the `h` and `s` helpers.
 */
export type HArgs = [
  string | Node,
  ...(Record<string, any> | null | undefined | Node | string | number | HArgs)[]
];

/**
 * The type of arguments of the `jcss` helper.
 */
export type JcssNode = {
  [attributeOrSelector: string]: string | number | JcssNode;
};

/**
 * The type of arguments of the `jcss` helper.
 */
export type JcssRoot = Record<string, JcssNode>;

/**
 * A helper for creating a chart based on a table (conf. <https://jackens.github.io/nnn/chartable/>).
 *
 * Options:
 * - `bottom`: bottom padding (for X axis labels)
 * - `gapX`: X axis spacing
 * - `gapY`: Y axis spacing
 * - `headerColumn`: flag indicating that `table` has a header column (with X axis labels)
 * - `headerRow`: flag indicating that `table` has a header row (with data series labels)
 * - `id`: chart id
 * - `left`: left padding (for data series labels)
 * - `maxY`: number of Y axis lines
 * - `right`: right padding (for data series labels)
 * - `singleScale`: flag to force single scale
 * - `table`: `HTMLTableElement` to extract data, data series labels and X axis labels
 * - `title`: chart title
 * - `top`: top padding (for the title)
 * - `xLabels`: X axis labels
 * - `zLabels`: data series labels
 * - `zxY`: chart data
 */
export function chartable(options?: {
  bottom?: number;
  gapX?: number;
  gapY?: number;
  headerColumn?: boolean;
  headerRow?: boolean;
  id?: string;
  left?: number;
  maxY?: number;
  right?: number;
  singleScale?: boolean;
  table?: HTMLTableElement;
  title?: string;
  top?: number;
  xLabels?: string[];
  zLabels?: string[];
  zxY?: number[][];
}): SVGSVGElement;

/**
 * A helper that returns information about the JavaScript engine.
 *
 * It returns an array of triples: `[«name», «prototype-name», «array-of-own-property-names»]`.
 */
export function engineInfo(): [string, string, string[]][];

/**
 * A helper that checks equality of the given arguments.
 */
export function eq(x: any, y: any): boolean;

/**
 * A generic helper for escaping `values` by given `escapeMap` (in *TemplateStrings* flavor).
 */
export function escape(escapeMap: EscapeMap, template: TemplateStringsArray, ...values: any[]): string;

/**
 * A generic helper for escaping `values` by given `escapeMap`.
 */
export function escapeValues(escapeMap: EscapeMap, values: any[]): string[];

/**
 * A helper that implements typographic corrections specific to Polish typography.
 */
export function fixTypography(node: Node): void;

/**
 * A convenient helper for getting values of nested objects.
 */
export function get(defaultValue: any, ref: any, ...keys: (string | number | symbol)[]): any;

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `HTMLElement`s (see also `s`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttribute`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttribute`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `HArgs` are passed to `h` and the results are appended to the element being created or modified.
 */
export function h<T extends keyof HTMLElementTagNameMap>(tag: T, ...args: HArgs[1][]): HTMLElementTagNameMap[T];
export function h<N extends Node>(node: N, ...args: HArgs[1][]): N;
export function h(...args: HArgs): Node;

/**
 * A replacement for the `in` operator (not to be confused with the `for-in` loop) that works properly.
 */
export function has(key: any, ref: any): boolean;

/**
 * A helper that checks if the given argument is of a certain type.
 */
export function is(type: BigIntConstructor, arg: any): arg is bigint;
export function is(type: BooleanConstructor, arg: any): arg is boolean;
export function is(type: NumberConstructor, arg: any): arg is number;
export function is(type: StringConstructor, arg: any): arg is string;
export function is(type: SymbolConstructor, arg: any): arg is symbol;
export function is(type: undefined, arg: any): arg is null | undefined;
export function is<T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;

/**
 * A simple CSS-in-JS helper.
 *
 * The `root` parameter provides a hierarchical description of CSS rules.
 *
 * - Keys of sub-objects whose values are NOT objects are treated as CSS attribute, and values are treated as values of those CSS attributes; the concatenation of keys of all parent objects is a CSS rule.
 * - All keys ignore the part starting with a splitter (default: `$$`) sign until the end of the key (e.g. `src$$1` → `src`, `@font-face$$1` → `@font-face`).
 * - In keys specifying CSS attribute, all uppercase letters are replaced by lowercase letters with an additional `-` character preceding them (e.g. `fontFamily` → `font-family`).
 * - Commas in keys that makes a CSS rule cause it to “split” and create separate rules for each part (e.g. `{div:{margin:1,'.a,.b,.c':{margin:2}}}` → `div{margin:1}div.a,div.b,div.c{margin:2}`).
 * - Top-level keys that begin with `@` are not concatenated with sub-object keys.
 */
export function jcss(root: JcssRoot, splitter?: string): string;

/**
 * `JSON.parse` with “JavaScript turned on”.
 *
 * Objects having *exactly* one property which is present in the `handlers` map, i.e. objects of the form:
 *
 * ```js
 * { "«handlerName»": [«params»] }
 * ```
 *
 * are replaced by the result of call
 *
 * ```js
 * handlers['«handlerName»'](...«params»)
 * ```
 */
export function jsOnParse(handlers: Record<string, Function>, text: string): any;

/**
 * Language translations helper.
 */
export function locale(
  locales: Record<string, Record<string, string | Record<string, string>>>,
  defaultLanguage: string,
  languages?: readonly string[]
): (text?: string, version?: string) => string;

/**
 * A generic helper for syntax highlighting (see also `nanolightJs`).
 */
export function nanolight(
  pattern: RegExp,
  highlighters: ((chunk: string, index: number) => HArgs[1])[],
  code: string
): HArgs[1][];

/**
 * A helper for highlighting JavaScript.
 */
export function nanolightJs(codeJs: string): HArgs[1][];

/**
 * A helper for choosing the correct singular and plural.
 */
export function plUral(singular: string, plural2: string, plural5: string, value: number): string;

/**
 * A lightweight [HyperScript](https://github.com/hyperhype/hyperscript)-style helper for creating and modifying `SVGElement`s (see also `h`).
 *
 * - The first argument of type `string` specifies the tag of the element to be created.
 * - The first argument of type `Node` specifies the element to be modified.
 * - All other arguments of type `Record<string, any>` are mappings of attributes and properties.
 *   Keys starting with `$` specify *properties* (without the leading `$`) to be set on the element being created or modified.
 *   (Note that `$` is not a valid attribute name character.)
 *   All other keys specify *attributes* to be set by `setAttributeNS`.
 *   An attribute equal to `false` causes the attribute to be removed by `removeAttributeNS`.
 * - All other arguments of type `null` or `undefined` are simply ignored.
 * - All other arguments of type `Node` are appended to the element being created or modified.
 * - All other arguments of type `string`/`number` are converted to `Text` nodes and appended to the element being created or modified.
 * - All other arguments of type `HArgs` are passed to `s` and the results are appended to the element being created or modified.
 */
export function s<T extends keyof SVGElementTagNameMap>(tag: T, ...args: HArgs[1][]): SVGElementTagNameMap[T];
export function s<N extends Node>(node: N, ...args: HArgs[1][]): N;
export function s(...args: HArgs): Node;

/**
 * A convenient helper for setting values of nested objects.
 */
export function set(value: any, ref: any, ...keys: (string | number | symbol)[]): void;

/**
 * A helper that generates a UUID v1 identifier (with a creation timestamp).
 *
 * - The optional `node` parameter should have the format `/^[0123456789abcdef]+$/`.
 *   Its value will be trimmed to last 12 characters and left padded with zeros.
 */
export function uuid1(options?: {
  date?: Date;
  node?: string;
}): string;
