/**
 * @typedef {Map<any, (value?: any) => string>} EscapeMap
 */

/**
 * @typedef {[
 *   string | Node,
 *   ...(Record<string, any> | null | undefined | Node | string | number | HArgs)[]
 * ]} HArgs
 */

/**
 * @typedef {{
 *   [attributeOrSelector: string]: string | number | JcssNode;
 * }} JcssNode
 */

/**
 * @typedef {Record<string, JcssNode>} JcssRoot
 */

const _COLORS = ['#e22', '#e73', '#fc3', '#ad4', '#4d9', '#3be', '#45d', '#c3e']

/**
 * @param {{
 *   bottom?: number;
 *   gapX?: number;
 *   gapY?: number;
 *   headerColumn?: boolean;
 *   headerRow?: boolean;
 *   id?: string;
 *   left?: number;
 *   maxY?: number;
 *   right?: number;
 *   singleScale?: boolean;
 *   table?: HTMLTableElement;
 *   title?: string;
 *   top?: number;
 *   xLabels?: string[];
 *   zLabels?: string[];
 *   zxY?: number[][];
 * }} options
 */
export const chartable = ({
  bottom = 50,
  gapX = 70,
  gapY = 30,
  headerColumn = false,
  headerRow = true,
  id,
  left = 200,
  maxY = 10,
  right = 200,
  singleScale = false,
  table,
  title,
  top = 70,
  xLabels = [],
  zLabels = [],
  zxY = []
} = {}) => {
  if (table != null) {
    const zxYNotPassed = !zxY.length
    const xLabelsNotPassed = !xLabels.length
    const zLabelsNotPassed = !zLabels.length

    table.querySelectorAll('tr').forEach((row, r) => row.querySelectorAll('td,th').forEach((col, c) => {
      const x = r - +headerRow
      const z = c - +headerColumn
      // @ts-expect-error
      const value = col.innerText

      if (x >= 0 && z >= 0 && zxYNotPassed) {
        set(parseFloat(value), zxY, z, x)
      } else if (x >= 0 && z < 0 && xLabelsNotPassed) {
        xLabels[x] = value
      } else if (x < 0 && z >= 0 && zLabelsNotPassed) {
        zLabels[z] = value
      }
    }))
  }

  let bestScales = [Infinity, -Infinity, Infinity, -Infinity, Infinity]

  const ranges = zxY.map(xY => {
    xY = xY.filter(y => !isNaN(y))
    return [Math.min(...xY), Math.max(...xY)]
  })

  if (singleScale) {
    bestScales = ranges.reduce(([totalMin, totalMax], [min, max]) =>
      [Math.min(totalMin, min), Math.max(totalMax, max)], bestScales)
  } else {
    ranges.forEach(([min1, max0]) => ranges.forEach(([min0, max1]) => {
      if (min0 >= min1 && max1 >= max0) {
        let min2 = Infinity
        let max2 = -Infinity

        ranges.forEach(([min, max]) => {
          if (min < min1 || max1 < max) {
            if (min2 > min) {
              min2 = min
            }
            if (max2 < max) {
              max2 = max
            }
          }
        })

        let cost = 0

        ranges.forEach(([min, max]) => {
          if (min < max) {
            cost += 1 - (max - min) / (min1 <= min && max <= max1 ? max1 - min1 : max2 - min2)
          }
        })

        if (bestScales[4] > cost) {
          bestScales = [min1, max1, min2, max2, cost]
        }
      }
    }))
  }

  const [min1, max1, min2, max2] = bestScales
  const maxX = zxY[0].length
  const xi = Array.from({ length: maxX }, (_, x) => x * gapX)
  const yi = Array.from({ length: maxY }, (_, y) => y * gapY)
  const w = gapX * (maxX - 1)
  const h = gapY * (maxY - 1)
  const svgW = left + w + right
  const svgH = top + h + bottom
  const /** @type {HArgs} */ graph = ['g', { transform: `translate(${left} ${top})` },
    ...yi.map(y => ['line', { x1: 0, x2: w, y1: y, y2: y, stroke: '#8888' }]),
    ...xi.map(x => ['line', { x1: x, x2: x, y1: 0, y2: h, stroke: '#8888' }])]
  const /** @type {HArgs} */ lColors = ['g', { class: 'chartable-z-colors', transform: 'translate(-80 0)' }]
  const /** @type {HArgs} */ lLabels = ['g', { class: 'chartable-z-labels', transform: 'translate(-95 0)' }]
  const /** @type {HArgs} */ rColors = ['g', { class: 'chartable-z-colors', transform: 'translate(80 0)' }]
  const /** @type {HArgs} */ rLabels = ['g', { class: 'chartable-z-labels', transform: 'translate(95 0)' }]
  const yLabel = (/** @type {number} */ min, /** @type {number} */ max, /** @type {number} */ y) =>
    (min + (max - min) * (maxY - 1 - y) / (maxY - 1)).toFixed(2).replace(/\.?0+$/, '')

  zxY.forEach((xa, z) => {
    const cls = `chartable-z-${z + 1}`
    const fill = _COLORS[z % _COLORS.length]
    let [min, max] = ranges[z]

    if (min1 <= min && max <= max1) {
      min = min1
      max = max1
      lColors.push(['circle', { class: cls, cx: 0, cy: gapY * (lColors.length - 2), r: 5, fill }])
      lLabels.push(['text',
        { x: 0, y: gapY * (lLabels.length - 2), 'text-anchor': 'end', 'alignment-baseline': 'middle' },
        zLabels[z]])
    } else {
      min = min2
      max = max2
      rColors.push(['circle', { class: cls, cx: 0, cy: gapY * (rColors.length - 2), r: 5, fill }])
      rLabels.push(['text',
        { x: 0, y: gapY * (rLabels.length - 2), 'text-anchor': 'start', 'alignment-baseline': 'middle' },
        zLabels[z]])
    }

    const y = xa.map(x => isNaN(x) ? x : h * (max - x) / (max - min))

    xi.forEach((x, i) => {
      if (!isNaN(y[i])) {
        graph.push(['circle', { class: cls, cx: x, cy: y[i], r: 5, fill }])

        if (!isNaN(y[i - 1])) {
          graph.push(['line', { class: cls, x1: x, x2: xi[i - 1], y1: y[i], y2: y[i - 1], stroke: fill }])
        }
      }
    })
  })

  return s('svg', {
    viewBox: `0 0 ${svgW} ${svgH}`,
    width: `${svgW}px`,
    height: `${svgH}px`,
    class: 'chartable'
  }, id != null ? { id } : null, title != null
    ? ['text',
        { class: 'chartable-title', x: svgW / 2, y: top / 2, 'text-anchor': 'middle', 'alignment-baseline': 'middle' },
        title]
    : null,
  graph,
  min1 < Infinity
    ? ['g', { class: 'chartable-left', transform: `translate(${left} ${top})` },
        ['g', { class: 'chartable-y-labels', transform: 'translate(-15 0)' },
          ...yi.map((y, i) => ['text',
            { x: 0, y, 'text-anchor': 'end', 'alignment-baseline': 'middle' },
            yLabel(min1, max1, i)])],
        lColors, lLabels]
    : null,
  min2 < Infinity
    ? ['g', { class: 'chartable-right', transform: `translate(${left + w} ${top})` },
        ['g', { class: 'chartable-y-labels', transform: 'translate(15 0)' },
          ...yi.map((y, i) => ['text',
            { x: 0, y, 'text-anchor': 'start', 'alignment-baseline': 'middle' },
            yLabel(min2, max2, i)])],
        rColors, rLabels]
    : null,
  xLabels.length > 0
    ? ['g', { transform: `translate(${left} ${top + h})` },
        ['g', { class: 'chartable-x-labels', transform: 'translate(0 15)' },
          ...xi.map((x, i) => ['text',
            { x, y: 0, 'text-anchor': 'middle', 'alignment-baseline': 'hanging' },
            xLabels[i]])]]
    : null
  )
}

export const engineInfo = () => {
  const /** @type {Set<Function>} */ fns = new Set()
  let ref = window

  while (ref != null) {
    for (const name of Object.getOwnPropertyNames(ref)) {
      // @ts-expect-error
      let fn = ref[name]

      if (is(Function, fn) && fn.toString?.()?.includes?.('[native code]')) {
        while (fn != null && !fns.has(fn)) {
          fns.add(fn)
          fn = Object.getPrototypeOf(fn)
        }
      }
    }

    ref = Object.getPrototypeOf(ref)
  }

  return Array.from(fns.values()).map(/** @returns {[string, string, string[]]} */ fn => [
    fn.name,
    Object.getPrototypeOf(fn)?.name ?? '',
    Object.getOwnPropertyNames(fn.prototype ?? Object.create(null)).sort()
  ]).sort((a, b) => -(a[0] < b[0]))
}

export const eq = /** @returns {boolean} */ (/** @type {any} */ x, /** @type {any} */ y) => {
  if (x === y) {
    return true
  }

  const xConstructor = x?.constructor

  if (xConstructor === y?.constructor) {
    if (xConstructor === Number) {
      return (isNaN(x) && isNaN(y)) || +x === +y
    }

    if (xConstructor === Date) {
      return +x === +y
    }

    if (xConstructor === String || xConstructor === RegExp) {
      return '' + x === '' + y
    }

    if (xConstructor === Array) {
      return x.length === y.length && x.every((/** @type {any} */ item, /** @type {number} */ index) => eq(item, y[index]))
    }

    if (xConstructor === Object) {
      const keysOfX = Object.keys(x)
      return keysOfX.length === Object.keys(y).length && keysOfX.every(key => eq(x[key], y[key]))
    }

    if (xConstructor === Set || xConstructor === Map) {
      if (x.size !== y.size) {
        return false
      }
      const xa = [...x]
      const ya = [...y]
      return xa.every(xv => ya.find(yv => eq(xv, yv)))
    }
  }

  return false
}

export const escape = (
  /** @type {EscapeMap}            */ escapeMap,
  /** @type {TemplateStringsArray} */ template,
  /** @type {any[]}                */ ...values
) => String.raw(template, ...escapeValues(escapeMap, values))

export const escapeValues = (
  /** @type {EscapeMap} */ escapeMap,
  /** @type {any[]}     */ values
) => values.map(value => (escapeMap.get(value?.constructor) ?? escapeMap.get(undefined))?.(value) ?? '')

const _TAGS_TO_SKIP = { IFRAME: 1, NOSCRIPT: 1, PRE: 1, SCRIPT: 1, STYLE: 1, TEXTAREA: 1 }

export const fixTypography = (/** @type {Node} */ node) => {
  const /** @type {Node[]} */ queue = [node]

  while (queue.length > 0) {
    const node0 = queue.shift()

    if (node0 instanceof Element) {
      for (let i = 0; i < node0.childNodes.length; ++i) {
        const childNode = node0.childNodes[i]

        if (childNode instanceof Text) {
          queue.push(childNode)
        } else if (childNode instanceof Element && !has(childNode.tagName, _TAGS_TO_SKIP)) {
          queue.push(childNode)
        }
      }
    } else if (node0 instanceof Text && node0.nodeValue != null && node0.nodeValue.trim() !== '') {
      let /** @type {Node} */ previousNode = node0

      node0.nodeValue.split(/(\s|\(|„)([aiouwz—]\s)/gi).forEach((chunk, i) => {
        i %= 3

        const currentNode = i === 2
          ? h('span', { style: 'white-space:nowrap' }, chunk)
          : i === 1
            ? document.createTextNode(chunk)
            : document.createTextNode(chunk.replace(/(\/(?=[^/\s])|\.(?=[^\s]))/g, '$1\u200B'))

        if (node0.parentNode != null) {
          node0.parentNode.insertBefore(currentNode, previousNode.nextSibling)
        }

        previousNode = currentNode
      })

      node0.parentNode?.removeChild(node0)
    }
  }
}

export const get = (
  /** @type {any} */ defaultValue,
  /** @type {any} */ ref,
  /** @type {(string | number | symbol)[]} */ ...keys
) => {
  for (const key of keys) {
    if (!has(key, ref)) {
      return defaultValue
    }

    ref = ref[key]
  }

  return ref
}

const /** @type {Record<string, string>} */ _NS = {
  xlink: 'http://www.w3.org/1999/xlink'
}

/**
 * @type {{
 *   (namespaceURI?: null | undefined): {
 *     <T extends keyof HTMLElementTagNameMap>(tag:  T, ...args: HArgs[1][]): HTMLElementTagNameMap[T];
 *     <N extends Node>                       (node: N, ...args: HArgs[1][]): N;
 *     (...args: HArgs): Node;
 *   };
 *   (namespaceURI: 'http://www.w3.org/2000/svg'): {
 *     <T extends keyof SVGElementTagNameMap> (tag:  T, ...args: HArgs[1][]): SVGElementTagNameMap[T];
 *     <N extends Node>                       (node: N, ...args: HArgs[1][]): N;
 *     (...args: HArgs): Node;
 *   };
 * }}
 */
const _h = (/** @type {string?=} */ namespaceURI) => {
  const createElement = namespaceURI == null
    ? (/** @type {string } */ tag) => document.createElement(tag)
    : (/** @type {string } */ tag) => document.createElementNS(namespaceURI, tag)

  const h = (/** @type {HArgs[0]} */ tagOrNode, /** @type {HArgs[1][]} */ ...args) => {
    const node = is(String, tagOrNode) ? createElement(tagOrNode) : tagOrNode

    for (const arg of args) {
      let child = null

      if (arg instanceof Node) {
        child = arg
      } else if (is(String, arg) || is(Number, arg)) {
        // @ts-expect-error
        child = new Text(arg)
      } else if (is(Array, arg)) {
        // @ts-expect-error
        child = h(...arg)
      } else if (arg != null) {
        for (const name in arg) {
          const value = arg[name]

          if (name[0] === '$') {
            const name1 = name.slice(1)

            if (is(Object, value)) {
              // @ts-expect-error
              node[name1] = node[name1] ?? {}
              // @ts-expect-error
              Object.assign(node[name1], value)
            } else {
              // @ts-expect-error
              node[name1] = value
            }
          } else if (node instanceof Element) {
            const indexOfColon = name.indexOf(':')

            if (indexOfColon >= 0) {
              const /** @type {string=} */ ns = _NS[name.slice(0, indexOfColon)]

              if (ns != null) {
                const basename = name.slice(indexOfColon + 1)

                if (value === true) {
                  node.setAttributeNS(ns, basename, '')
                } else if (value === false) {
                  node.removeAttributeNS(ns, basename)
                } else {
                  node.setAttributeNS(ns, basename, is(String, value) ? value : '' + value)
                }
              }
            } else {
              if (value === true) {
                node.setAttribute(name, '')
              } else if (value === false) {
                node.removeAttribute(name)
              } else {
                node.setAttribute(name, is(String, value) ? value : '' + value)
              }
            }
          }
        }
      }

      if (child != null) {
        node.appendChild(child)
      }
    }

    return node
  }

  return h
}

export const h = _h()

export const has = (/** @type {any} */ key, /** @type {any} */ ref) =>
  (is(String, key) || is(Number, key) || is(Symbol, key)) && Object.hasOwnProperty.call(ref ?? Object, key)

/**
 * @template {abstract new (...args: any[]) => any} T
 *
 * @type {{
 *   (type: BigIntConstructor,  arg: any): arg is bigint;
 *   (type: BooleanConstructor, arg: any): arg is boolean;
 *   (type: NumberConstructor,  arg: any): arg is number;
 *   (type: StringConstructor,  arg: any): arg is string;
 *   (type: SymbolConstructor,  arg: any): arg is symbol;
 *   (type: undefined,          arg: any): arg is undefined | null;
 *   <T extends abstract new (...args: any[]) => any>(type: T, arg: any): arg is InstanceType<T>;
 * }}
 *
 * @returns {arg is bigint | boolean | number | string | symbol | undefined | null | InstanceType<T>}
 */
export const is = (/** @type {T} */ type, /** @type {any} */ arg) => arg?.constructor === type

const _jcss = (
  /** @type {JcssNode} */ node,
  /** @type {string}   */ prefix,
  /** @type {string[]} */ result,
  /** @type {(text: string) => string} */ split
) => {
  const /** @type {[JcssNode | string[], string][]} */ queue = [[node, prefix]]

  while (queue.length) {
    const [style2, prefix2] = queue.shift() ?? []

    if (style2 == null || prefix2 == null) {
      continue
    }

    if (is(Array, style2)) {
      result.push(prefix2, prefix2 !== '' ? '{' : '', style2.join(';'), prefix2 !== '' ? '}' : '')
    } else {
      const /** @type {[JcssNode | string[], string][]} */ todo = []
      let /** @type {string[]} */ attributes = []
      let attributesPushed = false

      for (const key in style2) {
        const value = style2[key]

        if (is(String, value) || is(Number, value)) {
          if (!attributesPushed) {
            attributesPushed = true
            attributes = []
            todo.push([attributes, prefix2])
          }

          attributes.push(`${split(key).replace(/([A-Z])/g, (_, letter) => '-' + letter.toLowerCase())}:${value}`)
        } else {
          attributesPushed = false

          const /** @type {string[]} */ newPrefix = []
          const keySplitted = key.split(',')

          for (const prefixItem of prefix2.split(',')) {
            for (const keyItem of keySplitted) {
              newPrefix.push(prefixItem + keyItem)
            }
          }

          todo.push([value, newPrefix.join(',')])
        }
      }

      queue.unshift(...todo)
    }
  }
}

export const jcss = (/** @type {JcssRoot} */ root, splitter = '$$') => {
  const split = (/** @type {string} */ text) => text.split(splitter)[0]
  const /** @type {string[]} */ result = []

  for (const key in root) {
    const value = root[key]

    if (key[0] === '@') {
      result.push(split(key) + '{')
      _jcss(value, '', result, split)
      result.push('}')
    } else {
      _jcss(value, split(key), result, split)
    }
  }

  return result.join('')
}

export const jsOnParse = (
  /** @type {Record<string, Function>} */ handlers,
  /** @type {string}                   */ text
) => JSON.parse(text, (key, value) => {
  if (is(Object, value)) {
    let isSecondKey = false

    for (key in value) {
      if (isSecondKey) {
        return value
      }
      isSecondKey = true
    }

    if (has(key, handlers) && is(Array, value[key])) {
      return handlers[key](...value[key])
    }
  }

  return value
})

const _locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ language,
  /** @type {string} */ text,
  /** @type {string} */ version
) => {
  // @ts-expect-error
  const v = locales?.[language]?.[version]?.[text]

  if (is(String, v)) {
    return v
  }

  const t = locales?.[language]?.[text]

  return is(String, t) ? t : text
}

export const locale = (
  /** @type {Record<string, Record<string, string | Record<string, string>>>} */ locales,
  /** @type {string} */ defaultLanguage,
  languages = navigator.languages
) => {
  for (const language of languages) {
    if (has(language, locales)) {
      // @ts-expect-error
      return _locale.bind(0, locales, language)
    }
  }

  // @ts-expect-error
  return _locale.bind(0, locales, defaultLanguage)
}

export const nanolight = (
  /** @type {RegExp} */ pattern,
  /** @type {((chunk: string, index: number) => HArgs[1])[]} */ highlighters,
  /** @type {string} */ code
) => {
  const /** @type {HArgs[1][]} */ result = []

  code.split(pattern).forEach((chunk, index) => {
    index %= highlighters.length
    if (chunk != null) {
      result.push(highlighters[index](chunk, index))
    }
  })

  return result
}

// @ts-expect-error
export const nanolightJs = nanolight.bind(0,
  /('.*?'|".*?"|`[\s\S]*?`)|(\/\/.*?\n|\/\*[\s\S]*?\*\/)|(break|case|catch|const|continue|debugger|default|delete|do|else|eval|export\s+type|export|extends|false|finally|for|from|function|goto|if|import|in|instanceof|is|keyof|let|NaN|new|null|package|return|super|switch|this|throw|true|try|typeof|undefined|var|void|while|with|yield)(?!\w)|([<>=.?:&|!~*/%+-])|(0x[\dabcdef]+|0o[01234567]+|0b[01]+|\d+(?:\.[\d_]+)?(?:e[+-]?[\d_]+)?)|([$\w]+)(?=\()|([$\wąćęłńóśżźĄĆĘŁŃÓŚŻŹ]+)/,
  [
    chunk => chunk, // no match
    chunk => ['u', chunk], // string literals
    chunk => ['em', chunk], // comments
    chunk => ['b', chunk], // keywords
    chunk => ['b', chunk], // operators
    chunk => ['u', chunk], // number literals
    chunk => ['u', chunk], // function calls
    chunk => ['i', chunk] // literals
  ]
)

export const plUral = (
  /** @type {string} */ singular,
  /** @type {string} */ plural2,
  /** @type {string} */ plural5,
  /** @type {number} */ value
) => {
  const absValue = Math.abs(value)
  const absValueMod10 = absValue % 10

  return value === 1
    ? singular
    : absValue === 12 || absValue === 13 || absValue === 14
      ? plural5
      : absValueMod10 === 2 || absValueMod10 === 3 || absValueMod10 === 4
        ? plural2
        : plural5
}

export const s = _h('http://www.w3.org/2000/svg')

export const set = (
  /** @type {any} */ value,
  /** @type {any} */ ref,
  /** @type {(string | number | symbol)[]} */ ...keys
) => {
  const last = keys.length - 1

  if (last < 0) {
    return
  }

  for (let k = 0; k < last; ++k) {
    const key = keys[k]

    if (!has(key, ref)) {
      ref[key] = is(Number, keys[k + 1]) ? [] : {}
    }

    ref = ref[key]
  }

  ref[keys[last]] = value
}

const _ZEROS = '0'.repeat(16)
let _counter = 0

export const uuid1 = ({
  date = new Date(),
  node = Math.random().toString(16).slice(2)
} = {}) => {
  const time = _ZEROS + (10_000 * (+date + 12_219_292_800_000)).toString(16)

  _counter = (_counter + 1) & 0x3fff

  return time.slice(-8).concat(
    '-',
    time.slice(-12, -8),
    // @ts-expect-error
    -1,
    time.slice(-15, -12),
    '-',
    (8 | (_counter >> 12)).toString(16),
    (_ZEROS + (_counter & 0xfff).toString(16)).slice(-3),
    '-',
    (_ZEROS + node).slice(-12))
}
